# -*- python -*-
# $RCSfile: psoutput.py,v $
# $Revision: 1.21 $
# $Author: reida $
# $Date: 2004/09/14 21:02:33 $

# This software was produced by NIST, an agency of the U.S. government,
# and by statute is not subject to copyright in the United States.
# Recipients of this software assume all responsibilities associated
# with its operation, modification and maintenance. However, to
# facilitate maintenance we ask that before distributing modified
# versions of this software, you first contact the authors at
# oof_manager@ctcms.nist.gov. 

from common import color
from common.IO import colormap
from common.IO import outputdevice
from common.primitives import Rectangle
from SWIG.common.coord import Coord
from SWIG.common import lock
from common.primitives import iPoint
from common.primitives import Point
import types
import math
import weakref

# Notes about the generated postscript:
#
#   The image-drawing routines defined in the header take advantage of
# post-script features for writing bitmap images, and there are some
# relatively subtle requirements in using them.  These requirements
# are currently met by the various "write" commands, of course, but
# maintainers should be aware of them.
#
# Trailingimage:
#
# The basic "trailingimage" routine is used to draw OOFImage data in
# PostScript.  It is based on the PostScript level 2 command
# "colorimage' -- it locally redefines the coordinates so that (1,1)
# points to the upper-right corner of the physical-space coordinates
# (which may not be square), then pushes the width, height, and depth
# of the image onto the PostScript operand stack, then pushes a
# transformation matrix which maps the (pixels_x by pixels_y) image
# space to the (1,1) user-space, including the y-axis inversion and
# displacement.
#   The actual "trailingimage" function adds the data
# source to the stack, which in this case is a function which reads
# data from the current file, i.e. the file being interpreted.  This
# function is called by colorimage as many times as required to get
# enough data to fill in all the pixels of the image with the
# appropriate number of components.  It then pushes a "false"
# indicating that the image only has one source of data with
# components interleaved, and the number of components, before finally
# running "colorimage" itself.
#   Because of the nature of this function, the giant hexadecimal string
# following the "trailingimage" invocation is read not by the
# PostScript interpreter itself, but by the data-source function
# passed as an argument to colorimage.  Colorimage is smart enough to
# ignore invalid hex codes, so newlines, tabs, spaces, etc. are all
# allowed within this string, but it does *not* understand PostScript
# comments -- since comment text is likely to contain valid hex data
# (particularly the letters a-f), comments here will almost certainly
# cause errors.  Since the big hex string is automatically generated,
# there is actually little danger of this.
#
# 
# Colormaskimage:
#
#   The colormaskimage function uses a PostScript level-3 form of the
# "image" command to directly construct a partially transparent image
# from hex data.  At execution time, it is assumed that the width,
# height, and a vector representing the mask (transparent) color are
# already on the stack.  The only noteworthy command in the
# colormaskimage function is where it sets the colorspace to RGB --
# the default start-up color space for PostScript is DeviceGray, and
# this leads to hard-to-diagnose errors, as the "image" command relies
# on the colorspace for cues on how to interpret the hex data.  The
# data follow the command in the same way as for "trailingimage", and
# have the same constraints.
#   The mask color is expressed as integers, which are derived from
# the 0->1 floating point retrieval from an RGBColor.  In spite of
# the potential for round-off errors leading to the wrong integers,
# this is believed to be safe, since the integers in the stringimage
# itself are obtained by the similar means.

unitscale = {}
unitscale['points'] = 1
unitscale['inches'] = 72.00
unitscale['cm'] = 28.346456693

# Line width is always in points -- the factor expresses the width
# provided in points in the current user-coordinate-system units.
linewidth_factor = 1.0
linewidth_lock = lock.Lock()

def set_linewidth_factor(value):
    global linewidth_factor
    linewidth_lock.acquire()
    linewidth_factor = value
    linewidth_lock.release()

def get_linewidth_factor():
    global linewidth_factor
    linewidth_lock.acquire()
    res = linewidth_factor
    linewidth_lock.release()
    return res
    

class PSLineWidth:
    def __init__(self, w):
        self.width = w
    def write(self, file):
        file.write("%f setlinewidth\n" % (self.width*get_linewidth_factor()))

class PSColor:
    def __init__(self, color):
        self.color = color
    def write(self, file):
        file.write("%f %f %f setrgbcolor\n" % self.color.rgb())

class PSSegment:
    def __init__(self, segment):
        self.segment = segment
    def write(self, file):
        file.write("%f %f %f %f seg\n" % (self.segment.start()[0],
                                          self.segment.start()[1],
                                          self.segment.end()[0],
                                          self.segment.end()[1]))

class PSCurve:
    def __init__(self, curve):
        self.curve = curve
    def writepts(self, file):
        file.write("newpath\n")
        p0 = self.curve.points()[0]
        file.write("%f %f moveto\n" % (p0[0], p0[1]))
        for pt in self.curve.points()[1:]:
            file.write("%f %f lineto\n" % (pt[0], pt[1]))
    def write(self, file):
        self.writepts(file)
        file.write("stroke\n")

class PSPolygon(PSCurve):
    def write(self, file):
        self.writepts(file)
        file.write("closepath stroke\n")

class PSFilledPolygon(PSCurve):
    def write(self, file):
        self.writepts(file)
        file.write("closepath fill\n")

class PSTriangle(PSFilledPolygon):
    def __init__(self, center,size,angle):
	self.center=center
	self.size=size
	self.angle=angle
	PSFilledPolygon.__init__(self,self)
    def points(self):
	r=self.size*get_linewidth_factor()/math.sqrt(3)
	return [self.center+Point(*[r*trig(2*math.pi*n/3-self.angle) for trig in (math.sin,math.cos)]) for n in range(3)]

class PSCompoundPolygon:
    def __init__(self, polygons):
        self.polygons = polygons
    def write(self, file):
        file.write("newpath\n")
        for polygon in self.polygons:
            self.writepts(polygon,file)
        file.write("stroke\n")
    def writepts(self,polygon,file):
        p0 = polygon.points()[0]
        file.write("%f %f moveto\n" % (p0[0], p0[1]))
        for pt in polygon.points()[1:]:
            file.write("%f %f lineto\n" % (pt[0], pt[1]))
        file.write("%f %f lineto\n" % (p0[0], p0[1])) # explicitly close

class PSCompoundFilledPolygon(PSCompoundPolygon):
    def write(self, file):
        file.write("newpath\n")
        for polygon in self.polygons:
            self.writepts(polygon, file)
        file.write("fill\n")

class PSComment:
    def __init__(self, comment):
        self.comment = comment
    def write(self, file):
        file.write("% " + self.comment + "\n")

class PSImage:
    def __init__(self, image,offset,size):
        self.image = image
	self.offset=offset
	self.size=size    # "Physical" size.
        self.pixel_size = image.sizeInPixels()
    def write(self, file):
        file.write("gsave\n%f %f translate\n%f %f scale\n"%(
            self.offset[0],self.offset[1],self.size[0],self.size[1]))
        file.write("%d %d 8\n" % (self.pixel_size[0],
          self.pixel_size[1]))
        file.write("[%f 0 0 %f 0 %f]\n" % (self.pixel_size[0],
          -self.pixel_size[1],self.pixel_size[1]))
        file.write("trailingimage\n%s\ngrestore\n"%
            (self.image.stringimage().hexstringimage()));

class PSShapedImage:
    def __init__(self, image,offset,size,shapecolor):
        self.image = image
	self.offset=offset
	self.size=size
	self.shapecolor=shapecolor
        self.pixel_size = image.sizeInPixels()
    def write(self, file):
        file.write("gsave\n%f %f translate\n%f %f scale\n"%(
            self.offset[0],self.offset[1],self.size[0],self.size[1]))
        file.write("%d %d\n" % (self.pixel_size[0], self.pixel_size[1]))
        red = self.shapecolor.getRed()*255
        green = self.shapecolor.getGreen()*255
        blue = self.shapecolor.getBlue()*255
        file.write("[%d %d %d]\n" % (blue, green, red) )
        stringimage=self.image.stringimage()
        file.write("colormaskimage\n%s\n" %
            stringimage.hexstringimage() )
	    #this just works for monochromatic images:
##	    color=self.image.getFG()  # but this function is obsolete...
##	    file.write("%f %f %f setrgbcolor\nmonotransimage\n%s\n"%
##		(color.getRed(), color.getGreen(), color.getBlue(),
##                 stringimage.hexstringmask(self.shapecolor)))
        file.write("grestore\n")

class PSCircle:
    def __init__(self, center, radius, filled):
	self.center=center
	self.radius=radius
	self.filled=filled
    def _radius(self):
	return self.radius
    def write(self, file):
        file.write("%f %f %f 0 360 arc "%
                   (self.center[0],self.center[1],self._radius()))
        if self.filled:
            file.write("fill\n")
        else:
            file.write("stroke\n")

class PSDot(PSCircle):
    def __init__(self,center,radius):
	PSCircle.__init__(self,center,radius,1)
    def _radius(self):
	return self.radius*get_linewidth_factor()

class PSLayer(outputdevice.NullLayer):
    # PSLayer stores a *weak* reference to the device, because the
    # layers are stored in weak key dictionaries in the Display, keyed
    # by the device.  Keeping a reference to the device here would
    # break that mechanism.  TODO LATER: implement a less fragile
    # design.
    def __init__(self, device):
        self.device = weakref.ref(device)
    def hide(self):
        self.device().inhibit()

#####################################################################


class PSoutput(outputdevice.OutputDevice):
    def __init__(self,
                 filename,
                 margin=0.1,           # Fractional whitespace on page bdy.
                 pageheight=11.0,      # height of page
                 pagewidth=8.5,        # width of page
                 units='inches',       # units for the above
                 linewidthfactor=1.0,
                 format='portrait'):    # 'portrait' or 'landscape'

        outputdevice.OutputDevice.__init__(self)

        # list of objects to be plotted, including color, line width,
        # and font changes
        self.objlist = []
        
	self.clear()
        
        if not units in unitscale.keys():
            raise "PSoutput: unknown units"
        if format != 'landscape' and format != 'portrait':
            raise "PSoutput: unknown page format"

        self.file = open(filename, 'w')

        # Convert dimensions to points
        scalefactor = unitscale[units]
        self.pageh = pageheight * scalefactor
        self.pagew = pagewidth * scalefactor
        self.drawing_height = int(self.pageh/(1.0+2.0*margin))
        self.drawing_width = int(self.pagew/(1.0+2.0*margin))
        self.linewidthfactor = linewidthfactor
        
        self.format = format

        self.currentColor = color.black
        self._linecolor = color.black
        self._fillcolor = color.black
        self.colormap = colormap.GrayMap()

        # Not much else can be done until the bounding box is
        # established, so the header and definitions are written when
        # "show" is called, and the actual range is known.

    def expand_range(self, obj):
        if self.range == None:
            self.range = obj.enclosing_rectangle()
        else:
            self.range.swallow(obj)

    ####

    # We keep track of two colors, the line color and the fill color,
    # but PostScript only knows about one.  So before drawing lines or
    # filling regions, call check_linecolor or check_fillcolor, and
    # the PS color will be switched if necessary.

    def check_linecolor(self):
        if self.currentColor != self._linecolor:
            self.currentColor = self._linecolor
            self.objlist.append(PSColor(self.currentColor))

    def check_fillcolor(self):
        if self.currentColor != self._fillcolor:
            self.currentColor = self._fillcolor
            self.objlist.append(PSColor(self.currentColor))


    ###########################################
    ## Definitions of primitives
    ###########################################

    def clear(self):
        self.objlist = []
        self.range = None

    def show(self):
        if self.range != None and not self.file.closed:
            xmax = self.range.xmax()
            xmin = self.range.xmin()
            ymax = self.range.ymax()
            ymin = self.range.ymin()

            dx = xmax - xmin
            dy = ymax - ymin
            
            h_over_w = float(self.drawing_height)/float(self.drawing_width)
            w_over_h = 1.0/h_over_w
            if self.format == 'portrait':
                if (dy/dx) > h_over_w:
                    # Aspect ratio is skinnier than page, height-limited.
                    points_per_unit = float(self.drawing_height)/dy
                    self.height = self.drawing_height
                    self.width = self.height*(dx/dy)
                else:
                    # Width-limited.
                    points_per_unit = float(self.drawing_width)/dx
                    self.width = self.drawing_width
                    self.height = self.width*(dy/dx)
            elif self.format == 'landscape':
                if (dy/dx) > w_over_h:
                    # Height-limited, but "height" is along drawing_width.
                    points_per_unit = float(self.drawing_width)/dy
                    self.height = self.drawing_width
                    self.width = self.height*(dx/dy)
                else:
                    # Width-limited
                    points_per_unit = float(self.drawing_height)/dx
                    self.width = self.drawing_height
                    self.height = self.width*(dy/dx)
            # Unrecognized format is already checked for.
                    

            self.pointsize = 1.0/points_per_unit
            set_linewidth_factor(self.linewidthfactor*self.pointsize)

            # Bounding box.
            if self.format == 'portrait':
                bbxmin = int(0.5*(self.pagew - self.width))
                bbxmax = int(bbxmin + self.width)
                bbymin = int(0.5*(self.pageh - self.height))
                bbymax = int(bbymin + self.height)
            elif self.format == 'landscape':
                bbxmin = int(0.5*(self.pagew - self.height))
                bbxmax = int(bbxmin + self.height)
                bbymin = int(0.5*(self.pageh - self.width))
                bbymax = int(bbymin + self.width)

            # Write header, including bounding box
            self.file.write("%!PS-Adobe-3.0 EPSF-3.0\n")
            self.file.write("%%%%BoundingBox: %d %d %d %d\n" %
                            (bbxmin, bbymin, bbxmax, bbymax))
            self.file.write("%%Creator: OOF2\n")
            self.file.write("%%EndComments\n")
            self.file.write("%%EndProlog\n")
	    self.file.write("\n/picstr 3 string def\n");
	    self.file.write("%prints an image. Functions are defined for all image operations for safety. Since the image is read from currentfile, it is safer to have the image actually be the next thing to be read in the file.\n");
	    self.file.write("/trailingimage\n");
	    self.file.write("{\n\t{ currentfile picstr readhexstring pop }\n");
	    self.file.write("\tfalse 3\n\tcolorimage\n} def\n");
	    self.file.write(
		"""%takes a width, height, and transparent color on the stack and displays a color, transparent image, with appropriate pixels rendered as transparent.  After colormaskimage, the image follows.
/colormaskimage
{
    /DeviceRGB setcolorspace
    /transpcolor exch def
    /width exch def
    /height exch def
    <<
    /ImageType 4
    /MaskColor transpcolor
    /Width width
    /Height height
    /ImageMatrix [ width 0 0 height neg 0 height ]
    /DataSource { currentfile picstr readhexstring pop }
    /BitsPerComponent 8
    /Decode [ 0 1 0 1 0 1 ]
    >>
    image
} def

%displays a monochromatic transparent image. width and height should be on the stack. The image follows the call.
/monotransimage
{
    /height exch def
    /width exch def
    width height
    false
    [width 0 0 height neg 0 height]
    { currentfile picstr readhexstring pop }
    imagemask
} def
	    """)
            self.file.write("gsave\n")

            self.file.write("\n% set coordinate system\n")
            if self.format == 'landscape':
                self.file.write("90 rotate\n")
                self.file.write("0 -%d translate\n" % self.pagew)

            self.file.write("% move origin to corner of bounding box\n")
            if self.format == 'landscape':
                self.file.write("%e %e translate\n" %
                                (0.5*(self.pageh - self.width),
                                 0.5*(self.pagew - self.height)))
            else:                           # portrait
                self.file.write("%e %e translate\n" %
                                (0.5*(self.pagew - self.width),
                                 0.5*(self.pageh - self.height)))
            self.file.write("% set scaling to use natural units\n")
            self.file.write("%e %e scale\n" % (points_per_unit,
                                               points_per_unit))
            self.file.write(
                "% move origin so that ll corner of bb is (xmin, ymin)\n")
            self.file.write("%e %e translate\n" % (-xmin, -ymin))

            self.file.write("0 setlinewidth\n")

            # From here on, all postscript graphics is done using the
            # "physical" units of the microstructure.

            # Set clipping region
            self.file.write("\n% Set clipping region\n")
            self.file.write("newpath\n")
            self.file.write("%e %e moveto\n" % (xmin, ymin))
            self.file.write("%e %e lineto\n" % (xmax, ymin))
            self.file.write("%e %e lineto\n" % (xmax, ymax))
            self.file.write("%e %e lineto closepath clip\n" % (xmin, ymax))

            # Define macros
            self.file.write("/seg {newpath moveto lineto stroke} def\n")

            for obj in self.objlist:
                obj.write(self.file)
                
            self.file.write("showpage grestore\n")

        self.file.close()

    def append(self, obj):
        if not self.inhibited:
            self.objlist.append(obj)

    def begin_layer(self):
        self.inhibited = 0
        return PSLayer(self)

    def end_layer(self):
        self.inhibited = 0

    def inhibit(self):
        self.inhibited = 1
        
    def set_lineWidth(self, w):
        self.append(PSLineWidth(w))
	self.currentLineWidth=w

    def set_lineColor(self, x):
        # PSColor object is added to self.objlist only when the color
        # is actually used
        if type(x)==type(1.0):
            self._linecolor = self.colormap(x)
        else:
            self._linecolor = x

    def set_fillColor(self, x):
        if type(x)==type(1.0):
            self._fillcolor = self.colormap(x)
        else:
            self._fillcolor = x

    def draw_segment(self, segment):
        self.check_linecolor()
        self.expand_range(segment)
        self.append(PSSegment(segment))

    def draw_dot(self,dot):
        self.check_linecolor()
	self.append(PSDot(dot,self.currentLineWidth))
	r=self.currentLineWidth*self.linewidthfactor
	self.expand_range(Rectangle(Coord(dot[0]-r,dot[1]-r),Coord(dot[0]+r,dot[1]+r)))

    def draw_triangle(self,center,angle):
        self.check_linecolor()
	self.append(PSTriangle(center,self.currentLineWidth,angle))
	r=(self.currentLineWidth*self.linewidthfactor)/math.sqrt(3)
	self.expand_range(Rectangle(Coord(center[0]-r,center[1]-r),Coord(center[0]+r,center[1]+r)))

    def draw_curve(self, curve):
        self.check_linecolor()
        self.expand_range(curve)
        self.append(PSCurve(curve))

    def draw_polygon(self, polygon):
        self.check_linecolor()
        if type(polygon) == types.ListType:
            for pgon in polygon:
                self.expand_range(pgon)
            self.append(PSCompoundPolygon(polygon))
        else:
            self.expand_range(polygon)
            self.append(PSPolygon(polygon))

    def fill_polygon(self, polygon):
        self.check_fillcolor()
        if type(polygon) == types.ListType:
            for pgon in polygon:
                self.expand_range(pgon)
            self.append(PSCompoundFilledPolygon(polygon))
        else:
            self.expand_range(polygon)
            self.append(PSFilledPolygon(polygon))

    def draw_circle(self, center, radius):
        self.check_linecolor()
	self.append(PSCircle(center,radius,0))
	r=radius
	self.expand_range(Rectangle(Coord(center[0]-r,center[1]-r),Coord(center[0]+r,center[1]+r)))

    def fill_circle(self, center, radius):
        self.check_linecolor()
	self.append(PSCircle(center,radius,1))
	r=radius
	self.expand_range(Rectangle(Coord(center[0]-r,center[1]-r),Coord(center[0]+r,center[1]+r)))

    def draw_image(self,image,offset,size):
	if(image.sizeInPixels()==iPoint(0,0)):
	    return
        self.append(PSImage(image,offset,size))
	self.expand_range(Rectangle(offset,Coord(offset[0]+size[0],
	    offset[1]+size[1])))

    def draw_shaped_image(self,image,offset,size,shapecolor):
	if(image.sizeInPixels()==iPoint(0,0)):
	    return
        self.append(PSShapedImage(image,offset,size,shapecolor))
	self.expand_range(Rectangle(offset,Coord(offset[0]+size[0],
	    offset[1]+size[1])))

    def comment(self, remark):
        self.append(PSComment(remark))

