# -*- python -*-
# $RCSfile: skeletongroupmenu.py,v $
# $Revision: 1.21 $
# $Author: langer $
# $Date: 2004/09/22 20:46:00 $

# This software was produced by NIST, an agency of the U.S. government,
# and by statute is not subject to copyright in the United States.
# Recipients of this software assume all responsibilities associated
# with its operation, modification and maintenance. However, to
# facilitate maintenance we ask that before distributing modified
# versions of this software, you first contact the authors at
# oof_manager@ctcms.nist.gov. 

# This file builds the menu commands that operate the methods
# of the various selectable-groups, corresponding to methods
# of GenericGroupSet class in skeletongroups.py

from SWIG.common import ooferror
from common import debug
from common import utils
from common.IO import mainmenu
from common.IO import oofmenu
from common.IO import parameter
from common.IO import whoville
from common.IO import automatic
from common.IO import reporter
from engine import skeletongroups
from engine.IO import skeletongroupparams
import common.microstructure
import types

AutomaticNameParameter = parameter.AutomaticNameParameter
NodeGroupParameter = skeletongroupparams.NodeGroupParameter
SegmentGroupParameter = skeletongroupparams.SegmentGroupParameter
ElementGroupParameter = skeletongroupparams.ElementGroupParameter

nodegroupmenu = oofmenu.OOFMenuItem("NodeGroup", cli_only=1)
segmentgroupmenu = oofmenu.OOFMenuItem("SegmentGroup", cli_only=1)
elementgroupmenu = oofmenu.OOFMenuItem("ElementGroup", cli_only=1)

mainmenu.OOF.addItem(nodegroupmenu)
mainmenu.OOF.addItem(segmentgroupmenu)
mainmenu.OOF.addItem(elementgroupmenu)


########################################################################
# Generic callbacks usable by all classes of selectables.  Menu items
# must have a data item which is a string corresponding to the name
# of the selection set in the skeletoncontext, e.g. "nodesgroups",
# so that the menuitem callback can retrieve the right groupset.

# Make a new group name known.

def _new_group(menuitem, skeleton, name):
    skelc = whoville.getClass('Skeleton')[skeleton]
    groupset = getattr(skelc, menuitem.data)
    if groupset.isGroup(name):
        raise ooferror.ErrUserError("Group %s already exists." % group)
    else:
        groupset.addGroup(name)


# Remove a group.

def _remove_group(menuitem, skeleton, group):
    skelc = whoville.getClass('Skeleton')[skeleton]
    groupset = getattr(skelc, menuitem.data)
    if not groupset.isGroup(group):
        raise ooferror.ErrUserError("Group %s does not exist." % group)
    else:
        groupset.removeGroup(group)


# Copy an existing group.
def _copy_group(menuitem, skeleton, group, new_name):
    skelc = whoville.getClass('Skeleton')[skeleton]
    groupset = getattr(skelc, menuitem.data)
    if groupset.isGroup(new_name):
        raise ooferror.ErrUserError("Group %s already exists." % new_name)
    if groupset.isGroup(group):
        groupset.copyGroup(group, new_name)


# Add the current selection to the named group, creating
# the group if required.

def _add_selection_to_group(menuitem, skeleton, group):
    skelc = whoville.getClass('Skeleton')[skeleton]
    groupset = getattr(skelc, menuitem.data)
    if not groupset.isGroup(group):
        groupset.addGroup(group)
    groupset.addSelectionToGroup(group)

# Remove the currently-selected objects from the named group.
def _remove_selection_from_group(menuitem, skeleton, group):
    skelc = whoville.getClass('Skeleton')[skeleton]
    groupset = getattr(skelc, menuitem.data)
    # if not groupset.isGroup(group):
    #     groupset.removeGroup(group)
    groupset.removeSelectionFromGroup(group)

    
# Rename an existing group.
def _rename_group(menuitem, skeleton, group, new_name):
    skelc = whoville.getClass('Skeleton')[skeleton]
    groupset = getattr(skelc, menuitem.data)
    groupset.renameGroup(group, new_name)
##     if groupset.isGroup(new_name):
##         raise ooferror.ErrUserError("Group %s already exists." % new_name)
##     if groupset.isGroup(group):
##         groupset.renameGroup(group, new_name)
        

# Clear a group.

def _clear_group(menuitem, skeleton, group):
    skelc = whoville.getClass('Skeleton')[skeleton]
    groupset = getattr(skelc, menuitem.data)
    if not groupset.isGroup(group):
        raise ooferror.ErrUserError("Group %s does not exist." % group)
    groupset.clearGroup(group)


# Query a node group.
def _query_node_group(menuitem, skeleton, group):
    skelc = whoville.getClass('Skeleton')[skeleton]
    members = skelc.nodegroups.get_group(group)
    reporter.report(">>> ", len(members), " nodes")

def _query_sgmt_group(menuitem, skeleton, group):
    skelc = whoville.getClass('Skeleton')[skeleton]
    members = skelc.segmentgroups.get_group(group)
    reporter.report(">>> ", len(members), " segments")

def _query_elem_group(menuitem, skeleton, group):
    skelc = whoville.getClass('Skeleton')[skeleton]
    members = skelc.elementgroups.get_group(group)
    area = 0.0
    homog = 0.0
    for element in members:
        area += element.area()
        homog += element.homogeneity(skelc.getObject())
    if len(members):
        homog /= len(members)
    reporter.report(">>> ", len(members), " elements,", " area: ", area,
                    ", average homogeneity: ", homog)

#######################################

class GroupNameResolver:
    # Callable object to be used as the resolver for the
    # AutomaticNameParameter for group names.  Requires derived
    # classes that define getBasename().
    def __init__(self, attrname):
        # attrname is the name of the SkeletonContext attribute that
        # holds the groupset, eg, "nodegroups".
        self.attrname = attrname
    def __call__(self, param, startname):
        skelname = param.group['skeleton'].value
        skelcontext = whoville.getClass('Skeleton')[skelname]
        groupset = getattr(skelcontext, self.attrname)
        return utils.uniqueName(self.getBasename(param, startname),
                                groupset.allGroups())

class NewGroupNameResolver(GroupNameResolver):
    # Gets the default name from the constructor arguments.
    def __init__(self, basename, attrname):
        GroupNameResolver.__init__(self, attrname)
        # basename is the defaultname to be used for the groups.
        self.basename = basename
    def getBasename(self, param, startname):
        if param.automatic():
            return self.basename
        return startname

class CopyGroupNameResolver(GroupNameResolver):
    # Gets the default name from the object being copied (or renamed).
    def getBasename(self, param, startname):
        if param.automatic():
            return param.group['group'].value
        return startname
    
#######################################    
# Node-specific menu items:

new = oofmenu.OOFMenuItem("New_Group", cli_only=1,
                          callback=_new_group,
                          params=parameter.ParameterGroup(
    whoville.WhoParameter("skeleton", whoville.getClass('Skeleton')),
    AutomaticNameParameter("name", value=automatic.automatic,
                           resolver=NewGroupNameResolver("nodegroup",
                                                         "nodegroups"))
    ))

new.data = "nodegroups"

nodegroupmenu.addItem(new)

rename = oofmenu.OOFMenuItem("Rename_Group", cli_only=1,
                             callback=_rename_group,
                             params=parameter.ParameterGroup(
    whoville.WhoParameter("skeleton", whoville.getClass('Skeleton')),
    NodeGroupParameter("group"),
    parameter.StringParameter("new_name", "",
                              tip="New name for the nodegroup")
##     AutomaticNameParameter("new_name",
##                            resolver=CopyGroupNameResolver("nodegroups"),
##                            value=automatic.automatic)
    ))

rename.data = "nodegroups"

nodegroupmenu.addItem(rename)


copy = oofmenu.OOFMenuItem("Copy_Group", cli_only=1,
                           callback=_copy_group,
                           params=parameter.ParameterGroup(
    whoville.WhoParameter("skeleton", whoville.getClass('Skeleton')),
    NodeGroupParameter("group"),
    AutomaticNameParameter("new_name",
                           resolver=CopyGroupNameResolver("nodegroups"),
                           value=automatic.automatic))) 

copy.data = "nodegroups"

nodegroupmenu.addItem(copy)


remove = oofmenu.OOFMenuItem("Delete_Group", cli_only=1,
                             callback=_remove_group,
                             params=[
    whoville.WhoParameter("skeleton", whoville.getClass('Skeleton')),
    NodeGroupParameter("group")]) 

remove.data = "nodegroups"

nodegroupmenu.addItem(remove)



addselect = oofmenu.OOFMenuItem("Add_to_Group", cli_only=1,
                                callback=_add_selection_to_group,
                                params=[
    whoville.WhoParameter("skeleton", whoville.getClass('Skeleton')),
    NodeGroupParameter("group")]) 

addselect.data = "nodegroups"

nodegroupmenu.addItem(addselect)


removeselect = oofmenu.OOFMenuItem("Remove_from_Group", cli_only=1,
                                callback=_remove_selection_from_group,
                                params=[
    whoville.WhoParameter("skeleton", whoville.getClass('Skeleton')),
    NodeGroupParameter("group")]) 

removeselect.data = "nodegroups"

nodegroupmenu.addItem(removeselect)


clear = oofmenu.OOFMenuItem("Clear_Group", cli_only=1,
                            callback=_clear_group,
                            params=[
    whoville.WhoParameter("skeleton", whoville.getClass('Skeleton')),
    NodeGroupParameter("group")]) 

clear.data = "nodegroups"

nodegroupmenu.addItem(clear)

query = oofmenu.OOFMenuItem("Query_Group", cli_only=1,
                            callback=_query_node_group,
                            params=[
    whoville.WhoParameter("skeleton", whoville.getClass('Skeleton')),
    NodeGroupParameter("group")]) 

query.data = "nodegroups"

nodegroupmenu.addItem(query)

#########################################################
# Segments.

new = oofmenu.OOFMenuItem("New_Group", cli_only=1,
                          callback=_new_group,
                          params=parameter.ParameterGroup(
    whoville.WhoParameter("skeleton", whoville.getClass('Skeleton')),
    AutomaticNameParameter("name",
                           value=automatic.automatic,
                           resolver=NewGroupNameResolver("segmentgroup",
                                                         "segmentgroups"))
    ))

new.data = "segmentgroups"

segmentgroupmenu.addItem(new)


rename = oofmenu.OOFMenuItem("Rename_Group", cli_only=1,
                             callback=_rename_group,
                             params=parameter.ParameterGroup(
    whoville.WhoParameter("skeleton", whoville.getClass('Skeleton')),
    SegmentGroupParameter("group"),
    parameter.StringParameter("new_name", "",
                              tip="New name for the segmentgroup")
##     AutomaticNameParameter("new_name",
##                            resolver=CopyGroupNameResolver("segmentgroups"),
##                            value=automatic.automatic)
    )) 

rename.data = "segmentgroups"

segmentgroupmenu.addItem(rename)


copy = oofmenu.OOFMenuItem("Copy_Group", cli_only=1,
                           callback=_copy_group,
                           params=parameter.ParameterGroup(
    whoville.WhoParameter("skeleton", whoville.getClass('Skeleton')),
    SegmentGroupParameter("group"),
    AutomaticNameParameter("new_name",
                           resolver=CopyGroupNameResolver("segmentgroups"),
                           value=automatic.automatic))) 

copy.data = "segmentgroups"

segmentgroupmenu.addItem(copy)


remove = oofmenu.OOFMenuItem("Delete_Group", cli_only=1,
                             callback=_remove_group,
                             params=[
    whoville.WhoParameter("skeleton", whoville.getClass('Skeleton')),
    SegmentGroupParameter("group")]) 

remove.data = "segmentgroups"

segmentgroupmenu.addItem(remove)



addselect = oofmenu.OOFMenuItem("Add_to_Group", cli_only=1,
                                callback=_add_selection_to_group,
                                params=[
    whoville.WhoParameter("skeleton", whoville.getClass('Skeleton')),
    SegmentGroupParameter("group")]) 

addselect.data = "segmentgroups"

segmentgroupmenu.addItem(addselect)


removeselect = oofmenu.OOFMenuItem("Remove_from_Group", cli_only=1,
                                callback=_remove_selection_from_group,
                                params=[
    whoville.WhoParameter("skeleton", whoville.getClass('Skeleton')),
    SegmentGroupParameter("group")]) 

removeselect.data = "segmentgroups"

segmentgroupmenu.addItem(removeselect)


clear = oofmenu.OOFMenuItem("Clear_Group", cli_only=1,
                            callback=_clear_group,
                            params=[
    whoville.WhoParameter("skeleton", whoville.getClass('Skeleton')),
    SegmentGroupParameter("group")]) 

clear.data = "segmentgroups"

segmentgroupmenu.addItem(clear)


query = oofmenu.OOFMenuItem("Query_Group", cli_only=1,
                            callback=_query_sgmt_group,
                            params=[
    whoville.WhoParameter("skeleton", whoville.getClass('Skeleton')),
    SegmentGroupParameter("group")]) 

query.data = "segmentgroups"

segmentgroupmenu.addItem(query)

########################################################
# Elements

new = oofmenu.OOFMenuItem("New_Group", cli_only=1,
                          callback=_new_group,
                          params=parameter.ParameterGroup(
    whoville.WhoParameter("skeleton", whoville.getClass('Skeleton')),
    AutomaticNameParameter("name",
                           value=automatic.automatic,
                           resolver=NewGroupNameResolver("elementgroup",
                                                         "elementgroups"))
    )) 

new.data = "elementgroups"

elementgroupmenu.addItem(new)


rename = oofmenu.OOFMenuItem("Rename_Group", cli_only=1,
                             callback=_rename_group,
                             params=parameter.ParameterGroup(
    whoville.WhoParameter("skeleton", whoville.getClass('Skeleton')),
    ElementGroupParameter("group"),
    parameter.StringParameter("new_name", "",
                              tip="New name for the elementgroup")
##     AutomaticNameParameter("new_name",
##                            resolver=CopyGroupNameResolver("elementgroups"),
##                            value=automatic.automatic)
    )) 

rename.data = "elementgroups"

elementgroupmenu.addItem(rename)


copy = oofmenu.OOFMenuItem("Copy_Group", cli_only=1,
                           callback=_copy_group,
                           params=parameter.ParameterGroup(
    whoville.WhoParameter("skeleton", whoville.getClass('Skeleton')),
    ElementGroupParameter("group"),
    AutomaticNameParameter("new_name",
                           resolver=CopyGroupNameResolver("elementgroups"),
                           value=automatic.automatic))) 

copy.data = "elementgroups"

elementgroupmenu.addItem(copy)


remove = oofmenu.OOFMenuItem("Delete_Group", cli_only=1,
                             callback=_remove_group,
                             params=[
    whoville.WhoParameter("skeleton", whoville.getClass('Skeleton')),
    ElementGroupParameter("group")]) 

remove.data = "elementgroups"

elementgroupmenu.addItem(remove)



addselect = oofmenu.OOFMenuItem("Add_to_Group", cli_only=1,
                                callback=_add_selection_to_group,
                                params=[
    whoville.WhoParameter("skeleton", whoville.getClass('Skeleton')),
    ElementGroupParameter("group")]) 

addselect.data = "elementgroups"

elementgroupmenu.addItem(addselect)


removeselect = oofmenu.OOFMenuItem("Remove_from_Group", cli_only=1,
                                callback=_remove_selection_from_group,
                                params=[
    whoville.WhoParameter("skeleton", whoville.getClass('Skeleton')),
    ElementGroupParameter("group")]) 

removeselect.data = "elementgroups"

elementgroupmenu.addItem(removeselect)


clear = oofmenu.OOFMenuItem("Clear_Group", cli_only=1,
                            callback=_clear_group,
                            params=[
    whoville.WhoParameter("skeleton", whoville.getClass('Skeleton')),
    ElementGroupParameter("group")]) 

clear.data = "elementgroups"

elementgroupmenu.addItem(clear)

query = oofmenu.OOFMenuItem("Query_Group", cli_only=1,
                            callback=_query_elem_group,
                            params=[
    whoville.WhoParameter("skeleton", whoville.getClass('Skeleton')),
    ElementGroupParameter("group")]) 

query.data = "elementgroups"

elementgroupmenu.addItem(query)
